@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2021 Gentoo Authors
+ * Copyright 2005-2022 Gentoo Authors
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2005-2010 Ned Ludd        - <solar@gentoo.org>
@@ -22,7 +22,7 @@
 #include "xasprintf.h"
 #include "xregex.h"
 
-#define QDEPENDS_FLAGS "drpbQitF:S" COMMON_FLAGS
+#define QDEPENDS_FLAGS "drpbQitUF:SR" COMMON_FLAGS
 static struct option const qdepends_long_opts[] = {
 	{"depend",    no_argument, NULL, 'd'},
 	{"rdepend",   no_argument, NULL, 'r'},
@@ -31,8 +31,10 @@ static struct option const qdepends_long_opts[] = {
 	{"query",     no_argument, NULL, 'Q'},
 	{"installed", no_argument, NULL, 'i'},
 	{"tree",      no_argument, NULL, 't'},
+	{"use",       no_argument, NULL, 'U'},
 	{"format",     a_argument, NULL, 'F'},
 	{"pretty",    no_argument, NULL, 'S'},
+	{"resolve",   no_argument, NULL, 'R'},
 	COMMON_LONG_OPTS
 };
 static const char * const qdepends_opts_help[] = {
@@ -43,22 +45,26 @@ static const char * const qdepends_opts_help[] = {
 	"Query reverse deps",
 	"Search installed packages using VDB",
 	"Search available ebuilds in the tree",
+	"Apply profile USE-flags to conditional deps",
 	"Print matched atom using given format string",
 	"Pretty format specified depend strings",
+	"Resolve found dependencies to package versions",
 	COMMON_OPTS_HELP
 };
 #define qdepends_usage(ret) usage(ret, QDEPENDS_FLAGS, qdepends_long_opts, qdepends_opts_help, NULL, lookup_applet_idx("qdepends"))
 
 /* structures / types / etc ... */
 struct qdepends_opt_state {
-	unsigned char qmode;
-	array_t *atoms;
-	array_t *deps;
-	set *udeps;
-	char *depend;
-	size_t depend_len;
-	const char *format;
-	tree_ctx *vdb;
+	unsigned int  qmode;
+	array_t      *atoms;
+	array_t      *deps;
+	set          *udeps;
+	char         *depend;
+	size_t        depend_len;
+	const char   *format;
+	char          resolve:1;
+	tree_ctx     *vdb;
+	tree_ctx     *rtree;
 };
 
 #define QMODE_DEPEND     (1<<0)
@@ -68,6 +74,7 @@ struct qdepends_opt_state {
 #define QMODE_INSTALLED  (1<<5)
 #define QMODE_TREE       (1<<6)
 #define QMODE_REVERSE    (1<<7)
+#define QMODE_FILTERUSE  (1<<8)
 
 const char *depend_files[] = {  /* keep *DEPEND aligned with above defines */
 	/* 0 */ "DEPEND",
@@ -171,7 +178,14 @@ qdepends_results_cb(tree_pkg_ctx *pkg_ctx, void *priv)
 			continue;
 		}
 
-		if (state->qmode & QMODE_TREE && verbose) {
+		/* try and resolve expressions to real package atoms */
+		if (state->resolve)
+			dep_resolve_tree(dep_tree, state->rtree);
+
+		if (state->qmode & QMODE_TREE &&
+			!(state->qmode & QMODE_REVERSE) &&
+			verbose)
+		{
 			/* pull in flags in use if possible */
 			tree_cat_ctx *vcat =
 				tree_open_cat(state->vdb, pkg_ctx->cat_ctx->name);
@@ -185,26 +199,29 @@ qdepends_results_cb(tree_pkg_ctx *pkg_ctx, void *priv)
 							             tree_pkg_meta_get(vpkg, BDEPEND);
 					if (depstr != NULL) {
 						dep_node *dep_vdb = dep_grow_tree(depstr);
-						if (dep_vdb != NULL)
+						if (dep_vdb != NULL) {
 							dep_flatten_tree(dep_vdb, state->deps);
-						else
+							dep_burn_tree(dep_vdb);
+						} else {
 							warn("failed to parse VDB depstring from %s\n",
-									atom_to_string(datom));
+								 atom_to_string(datom));
+						}
 					}
 					tree_close_pkg(vpkg);
 				}
 				tree_close_cat(vcat);
 			}
 		} else {
-			dep_prune_use(dep_tree, ev_use);
+			if (state->qmode & QMODE_FILTERUSE)
+				dep_prune_use(dep_tree, ev_use);
 			dep_flatten_tree(dep_tree, state->deps);
 		}
 
 		if (verbose) {
 			if (state->qmode & QMODE_REVERSE) {
-				array_for_each(state->atoms, m, atom) {
-					array_for_each(state->deps, n, fatom) {
-						if (atom_compare(fatom, atom) == EQUAL) {
+				array_for_each(state->deps, m, atom) {
+					array_for_each(state->atoms, n, fatom) {
+						if (atom_compare(atom, fatom) == EQUAL) {
 							fatom = NULL;
 							break;
 						}
@@ -256,15 +273,17 @@ qdepends_results_cb(tree_pkg_ctx *pkg_ctx, void *priv)
 						snprintf(buf, sizeof(buf), "%s%s%s",
 								RED, atom_to_string(atom), NORM);
 						if (quiet < 2)
-							add_set_unique(buf, state->udeps, NULL);
+							state->udeps = add_set_unique(buf,
+														  state->udeps, NULL);
 					} else if (!quiet) {
-						add_set_unique(atom_to_string(atom),
-								state->udeps, NULL);
+						state->udeps = add_set_unique(atom_to_string(atom),
+													  state->udeps, NULL);
 					}
 				}
 			} else {
 				array_for_each(state->deps, m, atom)
-					add_set_unique(atom_to_string(atom), state->udeps, NULL);
+					state->udeps = add_set_unique(atom_to_string(atom),
+												  state->udeps, NULL);
 			}
 		}
 
@@ -292,12 +311,14 @@ int qdepends_main(int argc, char **argv)
 	DECLARE_ARRAY(atoms);
 	DECLARE_ARRAY(deps);
 	struct qdepends_opt_state state = {
-		.atoms = atoms,
-		.deps = deps,
-		.udeps = create_set(),
-		.qmode = 0,
-		.format = "%[CATEGORY]%[PF]",
-		.vdb = NULL,
+		.atoms   = atoms,
+		.deps    = deps,
+		.udeps   = create_set(),
+		.qmode   = 0,
+		.format  = "%[CATEGORY]%[PF]",
+		.resolve = false,
+		.vdb     = NULL,
+		.rtree   = NULL,
 	};
 	size_t i;
 	int ret;
@@ -317,7 +338,9 @@ int qdepends_main(int argc, char **argv)
 		case 'Q': state.qmode |= QMODE_REVERSE;   break;
 		case 'i': state.qmode |= QMODE_INSTALLED; break;
 		case 't': state.qmode |= QMODE_TREE;      break;
+		case 'U': state.qmode |= QMODE_FILTERUSE; break;
 		case 'S': do_pretty = true;               break;
+		case 'R': state.resolve = true;           break;
 		case 'F': state.format = optarg;          break;
 		}
 	}
@@ -334,7 +357,7 @@ int qdepends_main(int argc, char **argv)
 	if (!(state.qmode & QMODE_INSTALLED) && !(state.qmode & QMODE_TREE))
 		state.qmode |= QMODE_INSTALLED;
 
-	/* don't allow both installed and froim tree */
+	/* don't allow both installed and from tree */
 	if (state.qmode & QMODE_INSTALLED && state.qmode & QMODE_TREE) {
 		warn("-i and -t cannot be used together, dropping -i");
 		state.qmode &= ~QMODE_INSTALLED;
@@ -381,6 +404,8 @@ int qdepends_main(int argc, char **argv)
 		array_for_each(overlays, n, overlay) {
 			t = tree_open(portroot, overlay);
 			if (t != NULL) {
+				if (state.resolve)
+					state.rtree = tree_open(portroot, overlay);
 				if (!(state.qmode & QMODE_REVERSE) && array_cnt(atoms) > 0) {
 					array_for_each(atoms, i, atom) {
 						ret |= tree_foreach_pkg_sorted(t,
@@ -391,9 +416,13 @@ int qdepends_main(int argc, char **argv)
 							qdepends_results_cb, &state, NULL);
 				}
 				tree_close(t);
+				if (state.rtree)
+					tree_close(state.rtree);
 			}
 		}
-	} else {
+	} else {  /* INSTALLED */
+		if (state.resolve)
+			state.rtree = tree_open_vdb(portroot, portvdb);
 		if (!(state.qmode & QMODE_REVERSE) && array_cnt(atoms) > 0) {
 			array_for_each(atoms, i, atom) {
 				ret |= tree_foreach_pkg_fast(state.vdb,
@@ -403,6 +432,8 @@ int qdepends_main(int argc, char **argv)
 			ret |= tree_foreach_pkg_fast(state.vdb,
 					qdepends_results_cb, &state, NULL);
 		}
+		if (state.rtree)
+			tree_close(state.rtree);
 	}
 
 	if (state.vdb != NULL)
